
%{
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
//-----------------------------------------------------------------------------
//----Compile with -DP_VERBOSE=1 for verbose output.
#ifndef P_VERBOSE
#  define P_VERBOSE 0
#endif
int verbose = P_VERBOSE;

//----Compile with -DP_USERPROC=1 to #include p_user_proc.c. p_user_proc.c 
//----should #define P_ACT, P_BUILD, P_TOKEN, P_PRINT to different procedures 
//----from those below, and supply code.
#ifdef P_USERPROC
#  include "p_user_proc.c"
#else
#  define P_ACT(ss) if(verbose)printf("%7d %s\n",yylineno,ss);
#  define P_BUILD(sym,A,B,C,D,E,F,G,H,I,J) pBuildTree(sym,A,B,C,D,E,F,G,H,I,J)
#  define P_TOKEN(tok,symbolIndex) pToken(tok,symbolIndex)
#  define P_PRINT(ss) if(verbose){printf("\n\n");pPrintTree(ss,0);}
#endif

extern int yylineno;
extern int yychar;
extern char yytext[];

extern int tptp_store_size;
extern char* tptp_lval[];

#define MAX_CHILDREN 12
typedef struct pTreeNode * pTree;
struct pTreeNode {
    char* symbol; 
    int symbolIndex; 
    pTree children[MAX_CHILDREN+1];
};
//-----------------------------------------------------------------------------
int yyerror( char *s ) { 

    fprintf( stderr, "%s in line %d at item \"%s\".\n", s, yylineno, yytext); 
}
//-----------------------------------------------------------------------------
pTree pBuildTree(char* symbol,pTree A,pTree B,pTree C,pTree D,pTree E,pTree F, 
pTree G, pTree H, pTree I, pTree J) { 

    pTree ss = (pTree)calloc(1,sizeof(struct pTreeNode));

    ss->symbol = symbol;
    ss->symbolIndex = -1;
    ss->children[0] = A; 
    ss->children[1] = B; 
    ss->children[2] = C;
    ss->children[3] = D;
    ss->children[4] = E;
    ss->children[5] = F;
    ss->children[6] = G;
    ss->children[7] = H;
    ss->children[8] = I;
    ss->children[9] = J;
    ss->children[10] = NULL;

    return ss; 
}
//-----------------------------------------------------------------------------
pTree pToken(char* token, int symbolIndex) { 

    char pTokenBuf[8240];
    pTree ss;
    char* symbol = tptp_lval[symbolIndex];
    char* safeSym;

    strncpy(pTokenBuf, token, 39);
    strncat(pTokenBuf, symbol, 8193);
    safeSym = strdup(pTokenBuf);
    ss = pBuildTree(safeSym,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
    ss->symbolIndex = symbolIndex;

    return ss; 
}
//-----------------------------------------------------------------------------
void pPrintComments(int start, int depth) { 

    int d, j;
    char c1[4] = "%", c2[4] = "/*";

    j = start;
    while (tptp_lval[j] != NULL && (tptp_lval[j][0]==c1[0] || 
(tptp_lval[j][0]==c2[0] && tptp_lval[j][1]==c2[1]))) { 
        for (d=0; d<depth-1; d++) {
            printf("| ");
        }
        printf("%1d ",depth % 10);
        printf("%s\n",tptp_lval[j]);
        j = (j+1)%tptp_store_size; 
    }
    return; 
}
//-----------------------------------------------------------------------------
void pPrintTree(pTree ss, int depth) { 

//----pPrintIdx is where to find top-level comments to print before a sentence. 
//----yywrap() gets those after last sentence.
    static int pPrintIdx = 0;
    int i, d;

    if (pPrintIdx >= 0) { 
        pPrintComments(pPrintIdx, 0); 
        pPrintIdx = -1;
    }
    if (ss == NULL) {
        return;
    }
    for (d = 0; d < depth-1; d++) {
        printf("| ");
    }
    printf("%1d ",depth % 10);
    if (ss->children[0] == NULL) {
        printf("%s\n", ss->symbol);
    } else {
        printf("<%s>\n", ss->symbol);
    }
    if (strcmp(ss->symbol, "PERIOD .") == 0) {
        pPrintIdx = (ss->symbolIndex+1) % tptp_store_size;
    }
    if (ss->symbolIndex >= 0) {
        pPrintComments((ss->symbolIndex+1) % tptp_store_size, depth);
    }
    i = 0;
    while(ss->children[i] != NULL) {
        pPrintTree(ss->children[i],depth+1); 
        i++;
    }
    return; 
}
//-----------------------------------------------------------------------------
int yywrap(void) { 

    P_PRINT(NULL); 
    return 1; 
}
//-----------------------------------------------------------------------------
%}

%union {int ival; double dval; char* sval; void* pval;}
%start  TPTP_file
%token <ival> AMPERSAND
%token <ival> AT_SIGN
%token <ival> AT_SIGN_MINUS
%token <ival> AT_SIGN_PLUS
%token <ival> CARET
%token <ival> COLON
%token <ival> COLON_EQUALS
%token <ival> COMMA
%token <ival> EQUALS
%token <ival> EQUALS_GREATER
%token <ival> EXCLAMATION
%token <ival> EXCLAMATION_EQUALS
%token <ival> EXCLAMATION_EXCLAMATION
%token <ival> EXCLAMATION_GREATER
%token <ival> LBRKT
%token <ival> LESS_EQUALS
%token <ival> LESS_EQUALS_GREATER
%token <ival> LESS_TILDE_GREATER
%token <ival> LPAREN
%token <ival> MINUS
%token <ival> MINUS_MINUS_GREATER
%token <ival> PERIOD
%token <ival> QUESTION
%token <ival> QUESTION_QUESTION
%token <ival> QUESTION_STAR
%token <ival> RBRKT
%token <ival> RPAREN
%token <ival> STAR
%token <ival> TILDE
%token <ival> TILDE_AMPERSAND
%token <ival> TILDE_VLINE
%token <ival> VLINE
%token <ival> _DLR_cnf
%token <ival> _DLR_fof
%token <ival> _DLR_fot
%token <ival> _DLR_itef
%token <ival> _DLR_itetf
%token <ival> _DLR_itett
%token <ival> _DLR_tff
%token <ival> _DLR_thf
%token <ival> _LIT_cnf
%token <ival> _LIT_fof
%token <ival> _LIT_include
%token <ival> _LIT_tff
%token <ival> _LIT_thf
%token <ival> arrow
%token <ival> comment
%token <ival> comment_line
%token <ival> decimal
%token <ival> decimal_exponent
%token <ival> decimal_fraction
%token <ival> distinct_object
%token <ival> dollar_dollar_word
%token <ival> dollar_word
%token <ival> dot_decimal
%token <ival> integer
%token <ival> less_sign
%token <ival> lower_word
%token <ival> plus
%token <ival> positive_decimal
%token <ival> rational
%token <ival> real
%token <ival> signed_integer
%token <ival> signed_rational
%token <ival> signed_real
%token <ival> single_quoted
%token <ival> star
%token <ival> unrecognized
%token <ival> unsigned_integer
%token <ival> unsigned_rational
%token <ival> unsigned_real
%token <ival> upper_word
%token <ival> vline
%%

TPTP_file : null {}
                    | TPTP_file TPTP_input {}
                    ;

TPTP_input : annotated_formula {P_PRINT($<pval>$);}
                    | include {P_PRINT($<pval>$);}
                    ;

annotated_formula : thf_annotated {$<pval>$ = P_BUILD("annotated_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_annotated {$<pval>$ = P_BUILD("annotated_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_annotated {$<pval>$ = P_BUILD("annotated_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | cnf_annotated {$<pval>$ = P_BUILD("annotated_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_annotated : _LIT_thf LPAREN name COMMA formula_role COMMA thf_formula annotations RPAREN PERIOD {$<pval>$ = P_BUILD("thf_annotated", P_TOKEN("_LIT_thf ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, $<pval>8, P_TOKEN("RPAREN ", $<ival>9), P_TOKEN("PERIOD ", $<ival>10));}
                    ;

tff_annotated : _LIT_tff LPAREN name COMMA formula_role COMMA tff_formula annotations RPAREN PERIOD {$<pval>$ = P_BUILD("tff_annotated", P_TOKEN("_LIT_tff ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, $<pval>8, P_TOKEN("RPAREN ", $<ival>9), P_TOKEN("PERIOD ", $<ival>10));}
                    ;

fof_annotated : _LIT_fof LPAREN name COMMA formula_role COMMA fof_formula annotations RPAREN PERIOD {$<pval>$ = P_BUILD("fof_annotated", P_TOKEN("_LIT_fof ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, $<pval>8, P_TOKEN("RPAREN ", $<ival>9), P_TOKEN("PERIOD ", $<ival>10));}
                    ;

cnf_annotated : _LIT_cnf LPAREN name COMMA formula_role COMMA cnf_formula annotations RPAREN PERIOD {$<pval>$ = P_BUILD("cnf_annotated", P_TOKEN("_LIT_cnf ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, $<pval>8, P_TOKEN("RPAREN ", $<ival>9), P_TOKEN("PERIOD ", $<ival>10));}
                    ;

annotations : COMMA source optional_info {$<pval>$ = P_BUILD("annotations", P_TOKEN("COMMA ", $<ival>1), $<pval>2, $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | null {$<pval>$ = P_BUILD("annotations", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

formula_role : lower_word {$<pval>$ = P_BUILD("formula_role", P_TOKEN("lower_word ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_formula : thf_logic_formula {$<pval>$ = P_BUILD("thf_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_sequent {$<pval>$ = P_BUILD("thf_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_logic_formula : thf_binary_formula {$<pval>$ = P_BUILD("thf_logic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_unitary_formula {$<pval>$ = P_BUILD("thf_logic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_type_formula {$<pval>$ = P_BUILD("thf_logic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_subtype {$<pval>$ = P_BUILD("thf_logic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_binary_formula : thf_binary_pair {$<pval>$ = P_BUILD("thf_binary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_binary_tuple {$<pval>$ = P_BUILD("thf_binary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_binary_type {$<pval>$ = P_BUILD("thf_binary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_binary_pair : thf_unitary_formula thf_pair_connective thf_unitary_formula {$<pval>$ = P_BUILD("thf_binary_pair", $<pval>1, $<pval>2, $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_binary_tuple : thf_or_formula {$<pval>$ = P_BUILD("thf_binary_tuple", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_and_formula {$<pval>$ = P_BUILD("thf_binary_tuple", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_apply_formula {$<pval>$ = P_BUILD("thf_binary_tuple", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_or_formula : thf_unitary_formula VLINE thf_unitary_formula {$<pval>$ = P_BUILD("thf_or_formula", $<pval>1, P_TOKEN("VLINE ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_or_formula VLINE thf_unitary_formula {$<pval>$ = P_BUILD("thf_or_formula", $<pval>1, P_TOKEN("VLINE ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_and_formula : thf_unitary_formula AMPERSAND thf_unitary_formula {$<pval>$ = P_BUILD("thf_and_formula", $<pval>1, P_TOKEN("AMPERSAND ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_and_formula AMPERSAND thf_unitary_formula {$<pval>$ = P_BUILD("thf_and_formula", $<pval>1, P_TOKEN("AMPERSAND ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_apply_formula : thf_unitary_formula AT_SIGN thf_unitary_formula {$<pval>$ = P_BUILD("thf_apply_formula", $<pval>1, P_TOKEN("AT_SIGN ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_apply_formula AT_SIGN thf_unitary_formula {$<pval>$ = P_BUILD("thf_apply_formula", $<pval>1, P_TOKEN("AT_SIGN ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_unitary_formula : thf_quantified_formula {$<pval>$ = P_BUILD("thf_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_unary_formula {$<pval>$ = P_BUILD("thf_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_atom {$<pval>$ = P_BUILD("thf_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_tuple {$<pval>$ = P_BUILD("thf_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_letrec {$<pval>$ = P_BUILD("thf_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_conditional {$<pval>$ = P_BUILD("thf_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN thf_logic_formula RPAREN {$<pval>$ = P_BUILD("thf_unitary_formula", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_quantified_formula : thf_quantifier LBRKT thf_variable_list RBRKT COLON thf_unitary_formula {$<pval>$ = P_BUILD("thf_quantified_formula", $<pval>1, P_TOKEN("LBRKT ", $<ival>2), $<pval>3, P_TOKEN("RBRKT ", $<ival>4), P_TOKEN("COLON ", $<ival>5), $<pval>6,NULL,NULL,NULL,NULL);}
                    ;

thf_variable_list : thf_variable {$<pval>$ = P_BUILD("thf_variable_list", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_variable COMMA thf_variable_list {$<pval>$ = P_BUILD("thf_variable_list", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_variable : thf_typed_variable {$<pval>$ = P_BUILD("thf_variable", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable {$<pval>$ = P_BUILD("thf_variable", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_typed_variable : variable COLON thf_top_level_type {$<pval>$ = P_BUILD("thf_typed_variable", $<pval>1, P_TOKEN("COLON ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_unary_formula : thf_unary_connective LPAREN thf_logic_formula RPAREN {$<pval>$ = P_BUILD("thf_unary_formula", $<pval>1, P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_type_formula : thf_typeable_formula COLON thf_top_level_type {$<pval>$ = P_BUILD("thf_type_formula", $<pval>1, P_TOKEN("COLON ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_typeable_formula : thf_atom {$<pval>$ = P_BUILD("thf_typeable_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_tuple {$<pval>$ = P_BUILD("thf_typeable_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN thf_logic_formula RPAREN {$<pval>$ = P_BUILD("thf_typeable_formula", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_subtype : constant subtype_sign constant {$<pval>$ = P_BUILD("thf_subtype", $<pval>1, $<pval>2, $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_top_level_type : thf_logic_formula {$<pval>$ = P_BUILD("thf_top_level_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_unitary_type : thf_unitary_formula {$<pval>$ = P_BUILD("thf_unitary_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_binary_type : thf_mapping_type {$<pval>$ = P_BUILD("thf_binary_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_xprod_type {$<pval>$ = P_BUILD("thf_binary_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_union_type {$<pval>$ = P_BUILD("thf_binary_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_mapping_type : thf_unitary_type arrow thf_unitary_type {$<pval>$ = P_BUILD("thf_mapping_type", $<pval>1, P_TOKEN("arrow ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_unitary_type arrow thf_mapping_type {$<pval>$ = P_BUILD("thf_mapping_type", $<pval>1, P_TOKEN("arrow ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_xprod_type : thf_unitary_type STAR thf_unitary_type {$<pval>$ = P_BUILD("thf_xprod_type", $<pval>1, P_TOKEN("STAR ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_xprod_type STAR thf_unitary_type {$<pval>$ = P_BUILD("thf_xprod_type", $<pval>1, P_TOKEN("STAR ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_union_type : thf_unitary_type plus thf_unitary_type {$<pval>$ = P_BUILD("thf_union_type", $<pval>1, P_TOKEN("plus ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_union_type plus thf_unitary_type {$<pval>$ = P_BUILD("thf_union_type", $<pval>1, P_TOKEN("plus ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_atom : term {$<pval>$ = P_BUILD("thf_atom", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_conn_term {$<pval>$ = P_BUILD("thf_atom", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_tuple : LBRKT RBRKT {$<pval>$ = P_BUILD("thf_tuple", P_TOKEN("LBRKT ", $<ival>1), P_TOKEN("RBRKT ", $<ival>2),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LBRKT thf_tuple_list RBRKT {$<pval>$ = P_BUILD("thf_tuple", P_TOKEN("LBRKT ", $<ival>1), $<pval>2, P_TOKEN("RBRKT ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_tuple_list : thf_unitary_formula {$<pval>$ = P_BUILD("thf_tuple_list", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_unitary_formula COMMA thf_tuple_list {$<pval>$ = P_BUILD("thf_tuple_list", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_letrec : COLON_EQUALS LBRKT thf_letrec_list RBRKT COLON thf_unitary_formula {$<pval>$ = P_BUILD("thf_letrec", P_TOKEN("COLON_EQUALS ", $<ival>1), P_TOKEN("LBRKT ", $<ival>2), $<pval>3, P_TOKEN("RBRKT ", $<ival>4), P_TOKEN("COLON ", $<ival>5), $<pval>6,NULL,NULL,NULL,NULL);}
                    ;

thf_letrec_list : thf_defined_var {$<pval>$ = P_BUILD("thf_letrec_list", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_defined_var COMMA thf_letrec_list {$<pval>$ = P_BUILD("thf_letrec_list", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_defined_var : thf_variable COLON_EQUALS thf_logic_formula {$<pval>$ = P_BUILD("thf_defined_var", $<pval>1, P_TOKEN("COLON_EQUALS ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN thf_defined_var RPAREN {$<pval>$ = P_BUILD("thf_defined_var", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_conditional : _DLR_itef LPAREN thf_logic_formula COMMA thf_logic_formula COMMA thf_logic_formula RPAREN {$<pval>$ = P_BUILD("thf_conditional", P_TOKEN("_DLR_itef ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, P_TOKEN("RPAREN ", $<ival>8),NULL,NULL);}
                    ;

thf_sequent : thf_tuple MINUS_MINUS_GREATER thf_tuple {$<pval>$ = P_BUILD("thf_sequent", $<pval>1, P_TOKEN("MINUS_MINUS_GREATER ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_formula : tff_logic_formula {$<pval>$ = P_BUILD("tff_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_typed_atom {$<pval>$ = P_BUILD("tff_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_logic_formula : tff_binary_formula {$<pval>$ = P_BUILD("tff_logic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_unitary_formula {$<pval>$ = P_BUILD("tff_logic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_binary_formula : tff_binary_nonassoc {$<pval>$ = P_BUILD("tff_binary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_binary_assoc {$<pval>$ = P_BUILD("tff_binary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_binary_nonassoc : tff_unitary_formula binary_connective tff_unitary_formula {$<pval>$ = P_BUILD("tff_binary_nonassoc", $<pval>1, $<pval>2, $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_binary_assoc : tff_or_formula {$<pval>$ = P_BUILD("tff_binary_assoc", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_and_formula {$<pval>$ = P_BUILD("tff_binary_assoc", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_or_formula : tff_unitary_formula VLINE tff_unitary_formula {$<pval>$ = P_BUILD("tff_or_formula", $<pval>1, P_TOKEN("VLINE ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_or_formula VLINE tff_unitary_formula {$<pval>$ = P_BUILD("tff_or_formula", $<pval>1, P_TOKEN("VLINE ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_and_formula : tff_unitary_formula AMPERSAND tff_unitary_formula {$<pval>$ = P_BUILD("tff_and_formula", $<pval>1, P_TOKEN("AMPERSAND ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_and_formula AMPERSAND tff_unitary_formula {$<pval>$ = P_BUILD("tff_and_formula", $<pval>1, P_TOKEN("AMPERSAND ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_unitary_formula : tff_quantified_formula {$<pval>$ = P_BUILD("tff_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_unary_formula {$<pval>$ = P_BUILD("tff_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | atomic_formula {$<pval>$ = P_BUILD("tff_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_letrec {$<pval>$ = P_BUILD("tff_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable {$<pval>$ = P_BUILD("tff_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_conditional {$<pval>$ = P_BUILD("tff_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN tff_logic_formula RPAREN {$<pval>$ = P_BUILD("tff_unitary_formula", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_quantified_formula : fol_quantifier LBRKT tff_variable_list RBRKT COLON tff_unitary_formula {$<pval>$ = P_BUILD("tff_quantified_formula", $<pval>1, P_TOKEN("LBRKT ", $<ival>2), $<pval>3, P_TOKEN("RBRKT ", $<ival>4), P_TOKEN("COLON ", $<ival>5), $<pval>6,NULL,NULL,NULL,NULL);}
                    ;

tff_variable_list : tff_variable {$<pval>$ = P_BUILD("tff_variable_list", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_variable COMMA tff_variable_list {$<pval>$ = P_BUILD("tff_variable_list", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_variable : tff_typed_variable {$<pval>$ = P_BUILD("tff_variable", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable {$<pval>$ = P_BUILD("tff_variable", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_typed_variable : variable COLON tff_atomic_type {$<pval>$ = P_BUILD("tff_typed_variable", $<pval>1, P_TOKEN("COLON ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_unary_formula : unary_connective tff_unitary_formula {$<pval>$ = P_BUILD("tff_unary_formula", $<pval>1, $<pval>2,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fol_infix_unary {$<pval>$ = P_BUILD("tff_unary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_typed_atom : tff_untyped_atom COLON tff_top_level_type {$<pval>$ = P_BUILD("tff_typed_atom", $<pval>1, P_TOKEN("COLON ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN tff_typed_atom RPAREN {$<pval>$ = P_BUILD("tff_typed_atom", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_untyped_atom : functor {$<pval>$ = P_BUILD("tff_untyped_atom", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | system_functor {$<pval>$ = P_BUILD("tff_untyped_atom", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_top_level_type : tff_atomic_type {$<pval>$ = P_BUILD("tff_top_level_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_mapping_type {$<pval>$ = P_BUILD("tff_top_level_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_unitary_type : tff_atomic_type {$<pval>$ = P_BUILD("tff_unitary_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN tff_xprod_type RPAREN {$<pval>$ = P_BUILD("tff_unitary_type", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_atomic_type : atomic_word {$<pval>$ = P_BUILD("tff_atomic_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | defined_type {$<pval>$ = P_BUILD("tff_atomic_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_mapping_type : tff_unitary_type arrow tff_atomic_type {$<pval>$ = P_BUILD("tff_mapping_type", $<pval>1, P_TOKEN("arrow ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN tff_mapping_type RPAREN {$<pval>$ = P_BUILD("tff_mapping_type", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_xprod_type : tff_atomic_type STAR tff_atomic_type {$<pval>$ = P_BUILD("tff_xprod_type", $<pval>1, P_TOKEN("STAR ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_xprod_type STAR tff_atomic_type {$<pval>$ = P_BUILD("tff_xprod_type", $<pval>1, P_TOKEN("STAR ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN tff_xprod_type RPAREN {$<pval>$ = P_BUILD("tff_xprod_type", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_letrec : COLON_EQUALS LBRKT tff_letrec_list RBRKT COLON tff_unitary_formula {$<pval>$ = P_BUILD("tff_letrec", P_TOKEN("COLON_EQUALS ", $<ival>1), P_TOKEN("LBRKT ", $<ival>2), $<pval>3, P_TOKEN("RBRKT ", $<ival>4), P_TOKEN("COLON ", $<ival>5), $<pval>6,NULL,NULL,NULL,NULL);}
                    ;

tff_letrec_list : tff_defined_var {$<pval>$ = P_BUILD("tff_letrec_list", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | tff_defined_var COMMA tff_letrec_list {$<pval>$ = P_BUILD("tff_letrec_list", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_defined_var : variable COLON_EQUALS tff_logic_formula {$<pval>$ = P_BUILD("tff_defined_var", $<pval>1, P_TOKEN("COLON_EQUALS ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable COLON MINUS term {$<pval>$ = P_BUILD("tff_defined_var", $<pval>1, P_TOKEN("COLON ", $<ival>2), P_TOKEN("MINUS ", $<ival>3), $<pval>4,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN tff_defined_var RPAREN {$<pval>$ = P_BUILD("tff_defined_var", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

tff_conditional : _DLR_itef LPAREN tff_logic_formula COMMA tff_logic_formula COMMA tff_logic_formula RPAREN {$<pval>$ = P_BUILD("tff_conditional", P_TOKEN("_DLR_itef ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, P_TOKEN("RPAREN ", $<ival>8),NULL,NULL);}
                    ;

fof_formula : fof_logic_formula {$<pval>$ = P_BUILD("fof_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_logic_formula : fof_binary_formula {$<pval>$ = P_BUILD("fof_logic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_unitary_formula {$<pval>$ = P_BUILD("fof_logic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_binary_formula : fof_binary_nonassoc {$<pval>$ = P_BUILD("fof_binary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_binary_assoc {$<pval>$ = P_BUILD("fof_binary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_binary_nonassoc : fof_unitary_formula binary_connective fof_unitary_formula {$<pval>$ = P_BUILD("fof_binary_nonassoc", $<pval>1, $<pval>2, $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_binary_assoc : fof_or_formula {$<pval>$ = P_BUILD("fof_binary_assoc", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_and_formula {$<pval>$ = P_BUILD("fof_binary_assoc", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_or_formula : fof_unitary_formula VLINE fof_unitary_formula {$<pval>$ = P_BUILD("fof_or_formula", $<pval>1, P_TOKEN("VLINE ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_or_formula VLINE fof_unitary_formula {$<pval>$ = P_BUILD("fof_or_formula", $<pval>1, P_TOKEN("VLINE ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_and_formula : fof_unitary_formula AMPERSAND fof_unitary_formula {$<pval>$ = P_BUILD("fof_and_formula", $<pval>1, P_TOKEN("AMPERSAND ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_and_formula AMPERSAND fof_unitary_formula {$<pval>$ = P_BUILD("fof_and_formula", $<pval>1, P_TOKEN("AMPERSAND ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_unitary_formula : fof_quantified_formula {$<pval>$ = P_BUILD("fof_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_unary_formula {$<pval>$ = P_BUILD("fof_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | atomic_formula {$<pval>$ = P_BUILD("fof_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_letrec {$<pval>$ = P_BUILD("fof_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable {$<pval>$ = P_BUILD("fof_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_conditional {$<pval>$ = P_BUILD("fof_unitary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN fof_logic_formula RPAREN {$<pval>$ = P_BUILD("fof_unitary_formula", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_quantified_formula : fol_quantifier LBRKT fof_variable_list RBRKT COLON fof_unitary_formula {$<pval>$ = P_BUILD("fof_quantified_formula", $<pval>1, P_TOKEN("LBRKT ", $<ival>2), $<pval>3, P_TOKEN("RBRKT ", $<ival>4), P_TOKEN("COLON ", $<ival>5), $<pval>6,NULL,NULL,NULL,NULL);}
                    ;

fof_variable_list : variable {$<pval>$ = P_BUILD("fof_variable_list", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable COMMA fof_variable_list {$<pval>$ = P_BUILD("fof_variable_list", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_unary_formula : unary_connective fof_unitary_formula {$<pval>$ = P_BUILD("fof_unary_formula", $<pval>1, $<pval>2,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fol_infix_unary {$<pval>$ = P_BUILD("fof_unary_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_letrec : COLON_EQUALS LBRKT fof_letrec_list RBRKT COLON fof_unitary_formula {$<pval>$ = P_BUILD("fof_letrec", P_TOKEN("COLON_EQUALS ", $<ival>1), P_TOKEN("LBRKT ", $<ival>2), $<pval>3, P_TOKEN("RBRKT ", $<ival>4), P_TOKEN("COLON ", $<ival>5), $<pval>6,NULL,NULL,NULL,NULL);}
                    ;

fof_letrec_list : fof_defined_var {$<pval>$ = P_BUILD("fof_letrec_list", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fof_defined_var COMMA fof_letrec_list {$<pval>$ = P_BUILD("fof_letrec_list", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_defined_var : variable COLON_EQUALS fof_logic_formula {$<pval>$ = P_BUILD("fof_defined_var", $<pval>1, P_TOKEN("COLON_EQUALS ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable COLON MINUS term {$<pval>$ = P_BUILD("fof_defined_var", $<pval>1, P_TOKEN("COLON ", $<ival>2), P_TOKEN("MINUS ", $<ival>3), $<pval>4,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LPAREN fof_defined_var RPAREN {$<pval>$ = P_BUILD("fof_defined_var", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fof_conditional : _DLR_itef LPAREN fof_logic_formula COMMA fof_logic_formula COMMA fof_logic_formula RPAREN {$<pval>$ = P_BUILD("fof_conditional", P_TOKEN("_DLR_itef ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, P_TOKEN("RPAREN ", $<ival>8),NULL,NULL);}
                    ;

cnf_formula : LPAREN disjunction RPAREN {$<pval>$ = P_BUILD("cnf_formula", P_TOKEN("LPAREN ", $<ival>1), $<pval>2, P_TOKEN("RPAREN ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | disjunction {$<pval>$ = P_BUILD("cnf_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

disjunction : literal {$<pval>$ = P_BUILD("disjunction", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | disjunction VLINE literal {$<pval>$ = P_BUILD("disjunction", $<pval>1, P_TOKEN("VLINE ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

literal : atomic_formula {$<pval>$ = P_BUILD("literal", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | TILDE atomic_formula {$<pval>$ = P_BUILD("literal", P_TOKEN("TILDE ", $<ival>1), $<pval>2,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | fol_infix_unary {$<pval>$ = P_BUILD("literal", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_conn_term : thf_pair_connective {$<pval>$ = P_BUILD("thf_conn_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | assoc_connective {$<pval>$ = P_BUILD("thf_conn_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | thf_unary_connective {$<pval>$ = P_BUILD("thf_conn_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fol_infix_unary : term infix_inequality term {$<pval>$ = P_BUILD("fol_infix_unary", $<pval>1, $<pval>2, $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_quantifier : fol_quantifier {$<pval>$ = P_BUILD("thf_quantifier", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | CARET {$<pval>$ = P_BUILD("thf_quantifier", P_TOKEN("CARET ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | EXCLAMATION_GREATER {$<pval>$ = P_BUILD("thf_quantifier", P_TOKEN("EXCLAMATION_GREATER ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | QUESTION_STAR {$<pval>$ = P_BUILD("thf_quantifier", P_TOKEN("QUESTION_STAR ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | AT_SIGN_PLUS {$<pval>$ = P_BUILD("thf_quantifier", P_TOKEN("AT_SIGN_PLUS ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | AT_SIGN_MINUS {$<pval>$ = P_BUILD("thf_quantifier", P_TOKEN("AT_SIGN_MINUS ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_pair_connective : infix_equality {$<pval>$ = P_BUILD("thf_pair_connective", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | infix_inequality {$<pval>$ = P_BUILD("thf_pair_connective", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | binary_connective {$<pval>$ = P_BUILD("thf_pair_connective", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

thf_unary_connective : unary_connective {$<pval>$ = P_BUILD("thf_unary_connective", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | EXCLAMATION_EXCLAMATION {$<pval>$ = P_BUILD("thf_unary_connective", P_TOKEN("EXCLAMATION_EXCLAMATION ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | QUESTION_QUESTION {$<pval>$ = P_BUILD("thf_unary_connective", P_TOKEN("QUESTION_QUESTION ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

subtype_sign : less_sign less_sign {$<pval>$ = P_BUILD("subtype_sign", P_TOKEN("less_sign ", $<ival>1), P_TOKEN("less_sign ", $<ival>2),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

fol_quantifier : EXCLAMATION {$<pval>$ = P_BUILD("fol_quantifier", P_TOKEN("EXCLAMATION ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | QUESTION {$<pval>$ = P_BUILD("fol_quantifier", P_TOKEN("QUESTION ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

binary_connective : LESS_EQUALS_GREATER {$<pval>$ = P_BUILD("binary_connective", P_TOKEN("LESS_EQUALS_GREATER ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | EQUALS_GREATER {$<pval>$ = P_BUILD("binary_connective", P_TOKEN("EQUALS_GREATER ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LESS_EQUALS {$<pval>$ = P_BUILD("binary_connective", P_TOKEN("LESS_EQUALS ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LESS_TILDE_GREATER {$<pval>$ = P_BUILD("binary_connective", P_TOKEN("LESS_TILDE_GREATER ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | TILDE_VLINE {$<pval>$ = P_BUILD("binary_connective", P_TOKEN("TILDE_VLINE ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | TILDE_AMPERSAND {$<pval>$ = P_BUILD("binary_connective", P_TOKEN("TILDE_AMPERSAND ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

assoc_connective : VLINE {$<pval>$ = P_BUILD("assoc_connective", P_TOKEN("VLINE ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | AMPERSAND {$<pval>$ = P_BUILD("assoc_connective", P_TOKEN("AMPERSAND ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

unary_connective : TILDE {$<pval>$ = P_BUILD("unary_connective", P_TOKEN("TILDE ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_type : atomic_defined_word {$<pval>$ = P_BUILD("defined_type", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

atomic_formula : plain_atomic_formula {$<pval>$ = P_BUILD("atomic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | defined_atomic_formula {$<pval>$ = P_BUILD("atomic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | system_atomic_formula {$<pval>$ = P_BUILD("atomic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

plain_atomic_formula : plain_term {$<pval>$ = P_BUILD("plain_atomic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_atomic_formula : defined_plain_formula {$<pval>$ = P_BUILD("defined_atomic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | defined_infix_formula {$<pval>$ = P_BUILD("defined_atomic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_plain_formula : defined_plain_term {$<pval>$ = P_BUILD("defined_plain_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_infix_formula : term defined_infix_pred term {$<pval>$ = P_BUILD("defined_infix_formula", $<pval>1, $<pval>2, $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_infix_pred : infix_equality {$<pval>$ = P_BUILD("defined_infix_pred", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

infix_equality : EQUALS {$<pval>$ = P_BUILD("infix_equality", P_TOKEN("EQUALS ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

infix_inequality : EXCLAMATION_EQUALS {$<pval>$ = P_BUILD("infix_inequality", P_TOKEN("EXCLAMATION_EQUALS ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

system_atomic_formula : system_term {$<pval>$ = P_BUILD("system_atomic_formula", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

term : function_term {$<pval>$ = P_BUILD("term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable {$<pval>$ = P_BUILD("term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | conditional_term {$<pval>$ = P_BUILD("term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

function_term : plain_term {$<pval>$ = P_BUILD("function_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | defined_term {$<pval>$ = P_BUILD("function_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | system_term {$<pval>$ = P_BUILD("function_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

plain_term : constant {$<pval>$ = P_BUILD("plain_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | functor LPAREN arguments RPAREN {$<pval>$ = P_BUILD("plain_term", $<pval>1, P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

constant : functor {$<pval>$ = P_BUILD("constant", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

functor : atomic_word {$<pval>$ = P_BUILD("functor", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_term : defined_atom {$<pval>$ = P_BUILD("defined_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | defined_atomic_term {$<pval>$ = P_BUILD("defined_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_atom : number {$<pval>$ = P_BUILD("defined_atom", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | distinct_object {$<pval>$ = P_BUILD("defined_atom", P_TOKEN("distinct_object ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_atomic_term : defined_plain_term {$<pval>$ = P_BUILD("defined_atomic_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_plain_term : defined_constant {$<pval>$ = P_BUILD("defined_plain_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | defined_functor LPAREN arguments RPAREN {$<pval>$ = P_BUILD("defined_plain_term", $<pval>1, P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_constant : defined_functor {$<pval>$ = P_BUILD("defined_constant", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

defined_functor : atomic_defined_word {$<pval>$ = P_BUILD("defined_functor", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

system_term : system_constant {$<pval>$ = P_BUILD("system_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | system_functor LPAREN arguments RPAREN {$<pval>$ = P_BUILD("system_term", $<pval>1, P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

system_constant : system_functor {$<pval>$ = P_BUILD("system_constant", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

system_functor : atomic_system_word {$<pval>$ = P_BUILD("system_functor", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

variable : upper_word {$<pval>$ = P_BUILD("variable", P_TOKEN("upper_word ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

arguments : term {$<pval>$ = P_BUILD("arguments", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | term COMMA arguments {$<pval>$ = P_BUILD("arguments", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

conditional_term : _DLR_itett LPAREN tff_logic_formula COMMA term COMMA term RPAREN {$<pval>$ = P_BUILD("conditional_term", P_TOKEN("_DLR_itett ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, P_TOKEN("RPAREN ", $<ival>8),NULL,NULL);}
                    | _DLR_itetf LPAREN fof_logic_formula COMMA term COMMA term RPAREN {$<pval>$ = P_BUILD("conditional_term", P_TOKEN("_DLR_itetf ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("COMMA ", $<ival>4), $<pval>5, P_TOKEN("COMMA ", $<ival>6), $<pval>7, P_TOKEN("RPAREN ", $<ival>8),NULL,NULL);}
                    ;

source : general_term {$<pval>$ = P_BUILD("source", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

optional_info : COMMA useful_info {$<pval>$ = P_BUILD("optional_info", P_TOKEN("COMMA ", $<ival>1), $<pval>2,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | null {$<pval>$ = P_BUILD("optional_info", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

useful_info : general_list {$<pval>$ = P_BUILD("useful_info", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

include : _LIT_include LPAREN file_name formula_selection RPAREN PERIOD {$<pval>$ = P_BUILD("include", P_TOKEN("_LIT_include ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, $<pval>4, P_TOKEN("RPAREN ", $<ival>5), P_TOKEN("PERIOD ", $<ival>6),NULL,NULL,NULL,NULL);}
                    ;

formula_selection : COMMA LBRKT name_list RBRKT {$<pval>$ = P_BUILD("formula_selection", P_TOKEN("COMMA ", $<ival>1), P_TOKEN("LBRKT ", $<ival>2), $<pval>3, P_TOKEN("RBRKT ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    | null {$<pval>$ = P_BUILD("formula_selection", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

name_list : name {$<pval>$ = P_BUILD("name_list", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | name COMMA name_list {$<pval>$ = P_BUILD("name_list", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

general_term : general_data {$<pval>$ = P_BUILD("general_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | general_data COLON general_term {$<pval>$ = P_BUILD("general_term", $<pval>1, P_TOKEN("COLON ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | general_list {$<pval>$ = P_BUILD("general_term", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

general_data : atomic_word {$<pval>$ = P_BUILD("general_data", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | atomic_word LPAREN general_terms RPAREN {$<pval>$ = P_BUILD("general_data", $<pval>1, P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    | variable {$<pval>$ = P_BUILD("general_data", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | number {$<pval>$ = P_BUILD("general_data", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | distinct_object {$<pval>$ = P_BUILD("general_data", P_TOKEN("distinct_object ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | formula_data {$<pval>$ = P_BUILD("general_data", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

formula_data : _DLR_thf LPAREN thf_formula RPAREN {$<pval>$ = P_BUILD("formula_data", P_TOKEN("_DLR_thf ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    | _DLR_tff LPAREN tff_formula RPAREN {$<pval>$ = P_BUILD("formula_data", P_TOKEN("_DLR_tff ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    | _DLR_fof LPAREN fof_formula RPAREN {$<pval>$ = P_BUILD("formula_data", P_TOKEN("_DLR_fof ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    | _DLR_cnf LPAREN cnf_formula RPAREN {$<pval>$ = P_BUILD("formula_data", P_TOKEN("_DLR_cnf ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    | _DLR_fot LPAREN term RPAREN {$<pval>$ = P_BUILD("formula_data", P_TOKEN("_DLR_fot ", $<ival>1), P_TOKEN("LPAREN ", $<ival>2), $<pval>3, P_TOKEN("RPAREN ", $<ival>4),NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

general_list : LBRKT RBRKT {$<pval>$ = P_BUILD("general_list", P_TOKEN("LBRKT ", $<ival>1), P_TOKEN("RBRKT ", $<ival>2),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | LBRKT general_terms RBRKT {$<pval>$ = P_BUILD("general_list", P_TOKEN("LBRKT ", $<ival>1), $<pval>2, P_TOKEN("RBRKT ", $<ival>3),NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

general_terms : general_term {$<pval>$ = P_BUILD("general_terms", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | general_term COMMA general_terms {$<pval>$ = P_BUILD("general_terms", $<pval>1, P_TOKEN("COMMA ", $<ival>2), $<pval>3,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

name : atomic_word {$<pval>$ = P_BUILD("name", $<pval>1,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | unsigned_integer {$<pval>$ = P_BUILD("name", P_TOKEN("unsigned_integer ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

atomic_word : lower_word {$<pval>$ = P_BUILD("atomic_word", P_TOKEN("lower_word ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | single_quoted {$<pval>$ = P_BUILD("atomic_word", P_TOKEN("single_quoted ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

atomic_defined_word : dollar_word {$<pval>$ = P_BUILD("atomic_defined_word", P_TOKEN("dollar_word ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

atomic_system_word : dollar_dollar_word {$<pval>$ = P_BUILD("atomic_system_word", P_TOKEN("dollar_dollar_word ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

number : integer {$<pval>$ = P_BUILD("number", P_TOKEN("integer ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | rational {$<pval>$ = P_BUILD("number", P_TOKEN("rational ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    | real {$<pval>$ = P_BUILD("number", P_TOKEN("real ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

file_name : single_quoted {$<pval>$ = P_BUILD("file_name", P_TOKEN("single_quoted ", $<ival>1),NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

null : {$<pval>$ = P_BUILD("null",NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);}
                    ;

